unit Viewer;

// Specialized viewer form for TFile3DS data using plain OpenGL.
//
// 07-SEP-99 ml: texturing, extended graphic import support
// 06-SEP-99 ml: auto centering of meshs
// 04-SEP-99 ml: improved normal vector calculation
// 28-AUG-99 ml: first implementation
//
// (c) Copyright 1999, Dipl. Ing. Mike Lischke (public@lischke-online.de)

interface

uses
  Windows, OpenGL, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Geometry, Types3DS, File3DS, Menus, ComCtrls, StdCtrls, ExtCtrls;

type
  TGLColor = TVector4f;           // array[0..3] of Single

  PTexPoint = ^TTexPoint;
  TTexPoint = record
    S, T: TGLFLoat;
  end;

  PTexPointArray = ^TTexPointArray;
  TTexPointArray = array[0..0] of TTexPoint;

  PFaceGroup = ^TFaceGroup;
  TFaceGroup = record
    TextureIndex: Integer;        // index into material list
    IndexCount: Cardinal;
    Indices: PIntegerArray;
  end;

  PMaterial = ^TMaterial;
  TMaterial = record
    Ambient,
    Diffuse,
    Specular,
    Emission: TGLColor;
    Shininess: Integer;
    PolygonMode: TGLEnum;       // GL_LINE for wire frame, GL_FILL for all other types
    ShadeModel: TGLEnum;        // GL_FLAT for wire frame and flat shading type, GL_SMOOTH for all other types
    TwoSided: Boolean;
    Texture: String;            // name of a BMP, GIF, PCX, JPG, BW or TGA file to be used as texture
    TextureObject: Cardinal;    // the OpenGL texture object name
    WrapS, WrapT,               // texture wrapping (GL_CLAMP, GL_REPEAT)
    MinFilter,                  // minification filter (GL_NEAREST, GL_LINEAR, GL_NEAREST_MIPMAP_NEAREST etc.)
    MagFilter,                  // magnification filter (GL_NEAREST, GL_LINEAR)
    Mode: Cardinal;             // texture mode (GL_DECAL, GL_MODULATE etc.)
  end;

  PLightSource = ^TLightSource;
  TLightSource = record
    Enabled,
    IsSpot: Boolean;
    Position: TVector;
    Ambient,
    Diffuse,
    Specular: TGLColor;
    SpotDirection: TAffineVector;
    SpotExponent,
    SpotCutOff: Single;
  end;

  PMeshObject = ^TMeshObject;
  TMeshObject = record
    MeshName: String;
    ListName: Cardinal;
    Visible: Boolean;
    Count: Cardinal;            // vertex count, only needed for auto center calculation
    Vertices: PVectorArray;     // comprises all vertices of the object in no particular order
    Normals: PVectorArray;
    TexCoords: PTexPointArray;
    FaceGroups: TList;          // a list of face groups, each comprising a material description and
                                // a list of indices into the VerticesArray array
  end;

  TViewerForm = class(TForm)
    PropertiesMenu: TPopupMenu;
    BkgndItem: TMenuItem;
    NavPanel: TPanel;
    ScaleEdit: TEdit;
    ScaleUpDown: TUpDown;
    Label1: TLabel;
    Label2: TLabel;
    NavCenterItem: TMenuItem;
    Label3: TLabel;
    Bevel1: TBevel;
    Label4: TLabel;
    FLLabel: TLabel;
    Label5: TLabel;
    N1: TMenuItem;
    Label6: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure ScaleUpDownChanging(Sender: TObject; var AllowChange: Boolean);
    procedure ScaleEditKeyPress(Sender: TObject; var Key: Char);
    procedure NavCenterItemClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
  private
    // rendering values
    FRenderingContext: HGLRC;
    FViewport: TRectangle;
    FDPI: Cardinal;
    // background settings
    FBackground: TBackground3DS;
    FBackgroundTexture: Cardinal;

    // camera values
    FFocalLength: Cardinal;
    FDepthOfView,
    FNearPlane: Single;
    FCameraChanged: Boolean;
    FSceneMatrix: TMatrix;

    // mesh values
    FLists: TList;
    FMainList: Cardinal; // the display list which conatins all other display lists
    FCenter: TAffineVector; // for auto centering

    // materials
    FMaterials: TStringList;

    // light sources
    FLights: TStringList;
    FLightSourcesChanged: Boolean;

    // navigation
    FLastX,
    FLastY: Integer;         // last mouse down position to calculate rotation

    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure ApplyPerspective(Viewport: TRectangle; Width, Height: Integer; DPI: Cardinal);
    procedure ClearBuffersAndBackground;
    procedure SetupLights;
    procedure CheckOpenGLError;
    procedure ClearLists;
    procedure ConvertLightSources(const Objects: TObjectList);
    procedure ConvertMaterials(const Materials3DS: TMaterialList);
    function ConvertMeshData(const Version: TReleaseLevel; const Mesh3DS: PMesh3DS; ShowWarnings: Boolean): PMeshObject;
    procedure SetupTextures(ImagePath: String);
    function CreateTextureObject(DC: HDC; Bitmap: TBitmap; MinFilter, MagFilter: Cardinal): Cardinal;
    function LoadPictureFromFile(Picture: TPicture; ImagePath, FileName: String): Boolean;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  public
    procedure Use3DSData(const Reader: TFile3DS);
    procedure EnumColor(const ColorName: String);
    procedure ColorClick(Sender: TObject);
  end;

var
  ViewerForm: TViewerForm;

//----------------------------------------------------------------------------------------------------------------------

implementation

{$R *.DFM}

uses Const3DS, Math, GraphicEx, Main, FileCtrl, Progress;

type EOpenGLError = class(Exception);

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.WMEraseBkgnd(var Message: TWMEraseBkgnd);

// avoids clearing the background (causes flickering and speed penalty), OpenGL will use the entire
// window anyway

begin
  Message.Result := 1;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.FormCreate(Sender: TObject);

// initializes OpenGL, creates a context for rendering and activates it

var ColorDepth: Cardinal;
    I: Integer;
    
begin
  ColorDepth := GetDeviceCaps(Canvas.Handle, BITSPIXEL) * GetDeviceCaps(Canvas.Handle, PLANES);
  FRenderingContext := CreateRenderingContext(Canvas.Handle, [opDoubleBuffered], ColorDepth, 0);
  ActivateRenderingContext(Canvas.Handle, FRenderingContext);
  // set some default options and values in the context
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);
  if ColorDepth < 24 then glEnable(GL_DITHER)
                     else glDisable(GL_DITHER);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  with FBackground do
  begin
    BgndUsed := btUseSolidBgnd;
    glClearColor(0, 0, 0, 1);
  end;

  // determine resolution of display (needed to calculate correct camera values)
  FDPI := GetDeviceCaps(Canvas.Handle, LOGPIXELSX);

  // default values
  FFocalLength := 70;
  FDepthOfView := 10000;
  FCameraChanged := True;

  // other variables filled with values from 3DS file:
  // a list of display list which can directly be used during rendering
  FLists := TList.Create;

  FMaterials := TStringList.Create;
  FMaterials.Sorted := True;
  FMaterials.Duplicates := dupError;

  FLights := TStringList.Create;
  FLights.Sorted := True;
  FLights.Duplicates := dupError;

  // prepare remaining stuff:
  // prepare color menu
  GetColorValues(EnumColor);
  for I := 1 to BkgndItem.Count - 1 do
    if (I mod 16) = 0 then BkgndItem[I].Break := mbBarBreak;

  ClearLists;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.FormDestroy(Sender: TObject);

begin
  // this call will just clear all local lists
  Use3DSData(nil);
  DeactivateRenderingContext;
  // free the rendering context
  DestroyRenderingContext(FRenderingContext);

  // free local data
  FLists.Free;
  FMaterials.Free;
  FLights.Free;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ApplyPerspective(Viewport: TRectangle; Width, Height: Integer; DPI: Cardinal);

// implementation of a camera model using focal length

var Left, Right,
    Top, Bottom,
    zFar, MaxDim,
    Ratio: Double;

begin
  // determine biggest dimension and resolution (height or width)
  MaxDim := Width;
  if Height > MaxDim then MaxDim := Height;

  // calculate near plane distance and extensions;
  // Scene ratio is determined by the window ratio. The viewport is just a
  // specific part of the entire window and has therefore no influence on the
  // scene ratio. What we need to know, though, is the ratio between the window
  // borders (left, top, right and bottom) and the viewport borders.
  // Note: viewport.top is actually bottom, because the window (and viewport) origin
  // in OGL is the lower left corner 

  // calculate window/viewport ratio for right extent
  Ratio := (2 * Viewport.Width + 2 * Viewport.Left - Width) / Width;
  // calculate aspect ratio correct right value of the view frustum and take
  // the window/viewport ratio also into account
  Right := Ratio * Width / (2 * MaxDim);

  // the same goes here for the other three extents
  // left extent:
  Ratio := (Width - 2 * Viewport.Left) / Width;
  Left := -Ratio * Width / (2 * MaxDim);

  // top extent (keep in mind the origin is left lower corner):
  Ratio := (2 * Viewport.Height + 2 * Viewport.Top - Height) / Height;
  Top := Ratio * Height / (2 * MaxDim);

  // bottom extent:
  Ratio := (Height - 2 * Viewport.Top) / Height;
  Bottom := -Ratio * Height / (2 * MaxDim);

  FNearPlane := FFocalLength * 2 * DPI / (25.4 * MaxDim);
  zFar := FNearPlane + FDepthOfView;

  // finally create view frustum
  glFrustum(Left, Right, Bottom, Top, FNearPlane, zFar);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ClearBuffersAndBackground;

// Prepares OGL buffers for one paint cycle and clears the background of the window. If there's a
// defined background in the 3DS data then this is used else a default solid colored one.

var MidPos: Single;

begin
  case FBackground.BgndUsed of
    btUseSolidBgnd: // just clear background and depth buffer, color has already been set
      glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
    btUseVGradientBgnd, // create a vertical gradient or
    btUseBitmapBgnd:    // bitmapped background
      begin
        glDisable(GL_LIGHTING);
        glDisable(GL_DEPTH_TEST);

        glMatrixMode(GL_MODELVIEW);
        glPushMatrix;
        glLoadIdentity;
        glMatrixMode(GL_PROJECTION);
        glPushMatrix;
        glLoadIdentity;
        with FViewport do
        begin
          glOrtho(0, Width - 1, 0, Height - 1, 0, 100);

          // draw two quads to simulate gradient background
          glFrontFace(GL_CCW);
          if FBackground.BgndUsed = btUseBitmapBgnd then
          begin
            glEnable(GL_TEXTURE_2D);
            glBindTexture(GL_TEXTURE_2D, FBackgroundTexture);
            glBegin(GL_QUAD_STRIP);
              glTexCoord2f(0, 0);
              glVertex2f(0, 0);
              glTexCoord2f(1, 0);
              glVertex2f(Width - 1, 0);
              glTexCoord2f(0, 1);
              glVertex2f(0, Height - 1);
              glTexCoord2f(1, 1);
              glVertex2f(Width - 1, Height - 1);
            glEnd;
          end
          else
          begin
            glDisable(GL_TEXTURE_2D);
            with FBackground.VGradient do
            begin
              MidPos := Height * GradPercent;
              glBegin(GL_QUAD_STRIP);
                glColor3f(Bottom.R, Bottom.G, Bottom.B);
                glVertex2f(0, 0);
                glVertex2f(Width - 1, 0);
                glColor3f(Mid.R, Mid.G, Mid.B);
                glVertex2f(0, MidPos);
                glVertex2f(Width - 1, MidPos);
                glColor3f(Top.R, Top.G, Top.B);
                glVertex2f(0, Height - 1);
                glVertex2f(Width - 1, Height - 1);
              glEnd;
            end;
          end;
        end;
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix;
        glMatrixMode(GL_PROJECTION);
        glPopMatrix;

        glEnable(GL_LIGHTING);
        glEnable(GL_DEPTH_TEST);
        glClear(GL_DEPTH_BUFFER_BIT); // clear only depth buffer
      end;
  else
    // solid background for not yet implemented backgrounds
    glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

var SpotOff: Single = 180;

procedure TViewerForm.SetupLights;

// imitializes OpenGL light sources from our FLights list

const L1: TVector4f = (-3000, 3000, 3000, 1);
      L2: TVector4f = (5000, 10000, 5000, 1);

      A: TVector4f = (0.2, 0.2, 0.2, 1);
      D: TVector4f = (0.7, 0.7, 0.7, 1);
      S: TVector4f = (1, 1, 1, 1);

var I: Integer;
    MaxLights: Integer;

begin
  FLightSourcesChanged := False;
  // setup two default light sources if none are defined in the 3DS file
  if FLights.Count = 0 then
  begin
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_POSITION, @L1);
    glLightfv(GL_LIGHT0, GL_AMBIENT, @A);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, @D);
    glLightfv(GL_LIGHT0, GL_SPECULAR, @S);
    glLightfv(GL_LIGHT0, GL_SPOT_CUTOFF, @SpotOff);

    glEnable(GL_LIGHT1);
    glLightfv(GL_LIGHT1, GL_POSITION, @L2);
    glLightfv(GL_LIGHT1, GL_AMBIENT, @A);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, @D);
    glLightfv(GL_LIGHT1, GL_SPECULAR, @S);
    glLightfv(GL_LIGHT1, GL_SPOT_CUTOFF, @SpotOff);
  end
  else
  begin
    // determine maximum number of supported light sources (8 are guaranteed)
    glGetIntegerv(GL_MAX_LIGHTS, @MaxLights);
    // only access as many lightsources as supported by the OpenGL implentation
    if FLights.Count < MaxLights then MaxLights := FLights.Count;
    for I := 0 to MaxLights - 1 do
    with PLightSource(FLights.Objects[I])^ do
    begin
      if Enabled then glEnable(GL_LIGHT0 + I)
                 else glDisable(GL_LIGHT0 + I);
      glLightfv(GL_LIGHT0 + I, GL_POSITION, @Position);
      glLightfv(GL_LIGHT0 + I, GL_AMBIENT, @Ambient);
      glLightfv(GL_LIGHT0 + I, GL_DIFFUSE, @Diffuse);
      glLightfv(GL_LIGHT0 + I, GL_SPECULAR, @Specular);
      if IsSpot then
      begin
        glLightfv(GL_LIGHT0 + I, GL_SPOT_DIRECTION, @SpotDirection);
        glLightfv(GL_LIGHT0 + I, GL_SPOT_EXPONENT, @SpotExponent);
        glLightfv(GL_LIGHT0 + I, GL_SPOT_CUTOFF, @SpotCutOff);
      end
      else // switch off spot light property by setting the spot cut off to 180
        glLightfv(GL_LIGHT0 + I, GL_SPOT_CUTOFF, @SpotOff);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function RoundDownToPowerOf2(Value: Integer): Integer;

var LogTwo: Extended;

begin
  LogTwo := log2(Value);
  if Trunc(LogTwo) < LogTwo then Result := Trunc(Power(2, Trunc(LogTwo)))
                            else Result := Value;
end;

//----------------------------------------------------------------------------------------------------------------------

function RoundUpToPowerOf2(Value: Integer): Integer;

var LogTwo: Extended;

begin
  LogTwo := log2(Value);
  if Trunc(LogTwo) < LogTwo then Result := Trunc(Power(2, Trunc(LogTwo) + 1))
                            else Result := Value;
end;

//----------------------------------------------------------------------------------------------------------------------

function TViewerForm.CreateTextureObject(DC: HDC; Bitmap: TBitmap; MinFilter, MagFilter: Cardinal): Cardinal;

// creates a 2D texture object from the given bitmap which must have width and height as being power of 2

var J: Integer;
    Data: PByteArray;
    BMInfo: TBitmapInfo;
    ImageSize: Cardinal;
    Temp: Byte;

begin
  Result := 0;
  if (Bitmap.Height > 2) and (Bitmap.Width > 2) then
  begin
    glGenTextures(1, @Result);
    glBindTexture(GL_TEXTURE_2D, Result);

    // get color components
    with BMinfo.bmiHeader do
    begin
      // create description of the required image format
      FillChar(BMInfo, SizeOf(BMInfo), 0);
      biSize := SizeOf(TBitmapInfoHeader);
      biBitCount := 24;
      biWidth := Bitmap.Width;
      // I want a top-down bitmap. It seems that a positive value in biHeight will always flip
      // the image (even if it is already a bottom-up image).
      biHeight := Bitmap.Height;
      if Cardinal(Bitmap.ScanLine[0]) < Cardinal(Bitmap.ScanLine[1]) then biHeight := - biHeight;
      ImageSize := Bitmap.Width * Bitmap.Height;
      biPlanes := 1;
      biCompression := BI_RGB;
      GetMem(Data, ImageSize * 3);
      try
        // get the actual bits of the image
        GetDIBits(DC, Bitmap.Handle, 0, Bitmap.Height, Data, BMInfo, DIB_RGB_COLORS);

        // Now set the bits depending on the features supported by OpenGL.
        if GL_EXT_bgra then
          // BGR extension avoids color component swapping
          if (MinFilter = GL_NEAREST) or (MinFilter = GL_LINEAR)
            then glTexImage2d(GL_TEXTURE_2D, 0, 3, biWidth, Bitmap.Height, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, Data)
            else gluBuild2DMipmaps(GL_TEXTURE_2D, 3, biWidth, Bitmap.Height, GL_BGR_EXT, GL_UNSIGNED_BYTE, Data)
          else
          begin
            // No BGR support, so we must swap the color components manually.
            {$ifopt R+} {$define RangeCheck} {$R-} {$endif}
            for J := 0 to ImageSize - 1 do // swap blue with red to go from bgr to rgb
            begin
              Temp := Data[J * 3];
              Data[J * 3]:=Data[J * 3 + 2];
              Data[J * 3 + 2] := Temp;
            end;
            {$ifdef RangeCheck} {$undef RangeCheck} {$R+} {$endif}
            if (MinFilter = GL_NEAREST) or (MinFilter = GL_LINEAR)
              then glTexImage2d(GL_TEXTURE_2D, 0, 3, biWidth, Bitmap.Height, 0, GL_RGB, GL_UNSIGNED_BYTE, Data)
              else gluBuild2DMipmaps(GL_TEXTURE_2D, 3, biWidth, Bitmap.Height, GL_RGB, GL_UNSIGNED_BYTE, Data);
          end;
      finally
        FreeMem(Data);
      end;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

function TViewerForm.LoadPictureFromFile(Picture: TPicture; ImagePath, FileName: String): Boolean;

// Loads the given file into Picture.
// NOTE: If the freeware lib NViewLib.DLL is installed (system folder!) then GIF, PCX, JPG and TGA files are
//       supported as texture. Additionally, a few other formats can also be loaded (BMP, CEL, PIC, BW, and RGB)
//       without the need of this DLL.
//       JPG/JPEG images can also be loaded (if NViewLib isn't installed) by including the JPG unit from Delphi).
//       Since 3DS files contain no path of image files we use a small find strategy to get them. First the image is
//       searched where the 3DS file is stored, then subfolders of this location named "Textures" or "Maps".
//       A default path might be specified which is then searched through as very first location.

var
  DefaultPath: String;

begin
  Result := True;
  DefaultPath := MainForm.DefaultMapEdit.Text;
  if not DirectoryExists(DefaultPath) then DefaultPath := ''
                                      else
    if DefaultPath[Length(DefaultPath)] <> '\' then DefaultPath := DefaultPath + '\';

  try
    // load image data
    if FileExists(DefaultPath + FileName) then Picture.LoadFromFile(DefaultPath + FileName)
                                          else
      if FileExists(ImagePath + FileName) then Picture.LoadFromFile(ImagePath + FileName)
                                          else
        if FileExists(ImagePath + 'Textures\' + FileName) then Picture.LoadFromFile(ImagePath + 'Textures\' + FileName)
                                                          else
          if FileExists(ImagePath + 'Maps\' + FileName) then Picture.LoadFromFile(ImagePath + 'Maps\' + FileName)
                                                        else
          begin
            MessageDlg('Texture image not found (' + FileName + ')', mtInformation, [mbOK], 0);
            Result := False;
          end;
  except
    on E: Exception do
    begin
      Result := False;
      Application.ShowException(E);
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.SetupTextures(ImagePath: String);

// initializes texture objects from our FMaterials list

var Picture: TPicture;
    I: Integer;
    MemDC: HDC;
    TargetWidth,
    TargetHeight: Integer;

begin
  Picture := TPicture.Create;
  try
    MemDC := CreateCompatibleDC(0);
    for I := 0 to FMaterials.Count - 1 do
      with PMaterial(FMaterials.Objects[I])^ do
      begin
        if Texture = '' then Continue;
        if not LoadPictureFromFile(Picture, ImagePath, Texture) then Continue;
        if (Picture.Width = 0) or (Picture.Height = 0) then
        begin
          MessageDlg('Invalid texture image  (' + Texture + ')', mtInformation, [mbOK], 0);
          Continue;
        end;
        // check width and height being power of 2
        with Picture do
        begin
          TargetWidth := RoundDownToPowerOf2(Width);
          TargetHeight := RoundDownToPowerOf2(Height);
          if (TargetWidth <> Width) or (TargetHeight <> Height) then Stretch(TargetWidth, TargetHeight, sfLanczos3, 0, Bitmap);
          // create texture object
          TextureObject := CreateTextureObject(MemDC, Bitmap, MinFilter, MagFilter);
        end;
        if TextureObject = 0 then
        begin
          MessageDlg('Texture object could not be created (' + Texture + ')', mtInformation, [mbOK], 0);
          Continue;
        end;
      end;

    DeleteDC(MemDC);
  finally
    Picture.Free;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.FormPaint(Sender: TObject);

// This is the main paint routine. If there's no data assigned yet then an empty window is drawn else the data stored
// in the various lists. These lists must already be set up.

var Scale: Single;

begin
  // Note: We don't need to activate the rendering context here, as we did this in the OnShow event
  //       handler and we have setup the window to use a private DC (which is therefore never reset).
  ClearBuffersAndBackground;

  if FCameraChanged then             
  begin
    FCameraChanged := False;
    // set up projection 
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity;
    if NavPanel.Visible then
      ApplyPerspective(FViewport, ClientWidth - NavPanel.Width, ClientHeight, FDPI)
                        else
      ApplyPerspective(FViewport, Width, ClientHeight, FDPI);
  end;

  // initialize transformation pipeline
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;

  // lights
  Scale := ScaleUpDown.Position / 1000;
  if FLightSourcesChanged then
  begin
    glScalef(Scale, Scale, Scale);
    SetupLights;
    glLoadIdentity;
  end;

  // camera
  gluLookAt(0, 0, 50,
            0, 0, -10,
            0, 1, 0);

  // objects
  if Abs(1 - Scale) > 0.1 then
  begin
    glEnable(GL_NORMALIZE);  // this will half the frame rate but is necessary for changing overall
                             // scale on the fly and still get the correct lighting
    glScalef(Scale, Scale, Scale);
  end
  else glDisable(GL_NORMALIZE);
  glMultMatrixf(@FSceneMatrix);
  if FMainList <> 0 then glCallList(FMainList);

  // copy back buffer to front
  SwapBuffers(Canvas.Handle);
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.FormResize(Sender: TObject);

// handle form resizing (viewport must be adjusted)

var W: Integer;

begin
  // specify a viewport
  W := ClientWidth;
  if NavPanel.Visible then Dec(W, NavPanel.Width);
  glViewport(0, 0, W, Height);
  with FViewport do
  begin
    Left := 0;
    Top := 0;
    Width := W;
    Height := ClientHeight;
  end;
  FCameraChanged := True;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ConvertMaterials(const Materials3DS: TMaterialList);

// Builds an internal list with values which can directly sent to OpenGL. If no materials are defined then
// a default material is used.

const DefaultMaterial: TMaterial =
        (Ambient: (0.1, 0.1, 0.1, 1);
         Diffuse: (0.5, 0.5, 0.5, 1);
         Specular: (1, 1, 1, 1);
         Emission: (0, 0, 0, 1);
         Shininess: 128;
         PolygonMode: GL_FILL;
         ShadeModel: GL_SMOOTH;
         TwoSided: False;
         );

var I: Integer;
    Material: PMaterial;

begin
  if Materials3DS.Count = 0 then
  begin
    // Instead of directly adding the constant record to the list I use a dynamic variable here,
    // else I would have to undertake effort to determine whether to free the data or not in the
    // ClearLists method.
    Material := AllocMem(SizeOf(TMaterial));
    Material^ := DefaultMaterial;
    FMaterials.AddObject('Default', Pointer(Material));
  end
  else
    for I := 0 to Materials3DS.Count - 1 do
    begin
      Material := AllocMem(SizeOf(TMaterial));
      with Materials3DS[I]^ do
      begin
        Material.Ambient := MakeVector([Ambient.R, Ambient.G, Ambient.B, Abs(Transparency)]);
        Material.Diffuse := MakeVector([Diffuse.R, Diffuse.G, Diffuse.B, Abs(Transparency)]);
        Material.Specular := MakeVector([Specular.R, Specular.G, Specular.B, Abs(Transparency)]);
        Material.Shininess := Round(Shininess * 128);
        { don't know how to apply the percentage emission value to be a emission color
        if SelFillum then
        begin
          Material.Emission := Material.Diffuse;
          VectorScale(Material.Emission, SelfIllumPct);
        end;}

        // shading must be expressed in several OpenGL flags
        Material.PolygonMode := GL_LINE;
        Material.ShadeModel := GL_FLAT;
        if Shading > stWire then Material.PolygonMode := GL_FILL;
        // Gouraud, Phong and Metal shading are all expressed as Gouraud shading since this is
        // the only smoothing type OpenGL can render
        if Shading > stFlat then Material.ShadeModel := GL_SMOOTH;
        Material.TwoSided := TwoSided;

        // currently default values for texturing
        Material.WrapS := GL_REPEAT;
        Material.WrapT := GL_REPEAT;
        Material.MinFilter := GL_NEAREST;
        Material.MagFilter := GL_NEAREST;
        // also this is just a guess, since I haven't found any description when to use
        // decal and when modulated texturing (might well be that we even need GL_BLEND
        // here to implement the texture strength value)
        if (Shininess = 0) {or (VectorLength(Material.Diffuse) = 0)} then
          Material.Mode := GL_DECAL
                                                                   else
          Material.Mode := GL_MODULATE;
        Material.Texture := Texture.Map.Name;

        FMaterials.AddObject(Name, Pointer(Material));
      end;
    end;
end;

//----------------------------------------------------------------------------------------------------------------------

{$ifopt R+} {$define RangeCheck} {$R-} {$endif}
function TViewerForm.ConvertMeshData(const Version: TReleaseLevel; const Mesh3DS: PMesh3DS; ShowWarnings: Boolean): PMeshObject;

// Converts the vertex and texture data of the given 3DS mesh into a structure which can
// directly be send to OpenGL.
// Note: Passing array references to the local assembler routines is NOT redundant as Delphi
//       compiles wrong code if a variable of this function (ConvertMeshData) is accessed by a
//       local asm function. The calculation turns out to be as would there be no stack frame for
//       the local asm routines, but it *is* there.

type
  TSmoothIndexEntry = array[0..31] of Cardinal;

  PSmoothIndexArray = ^TSmoothIndexArray;
  TSmoothIndexArray = array[0..0] of TSmoothIndexEntry;

var
  Marker: PByteArray;
  CurrentVertexCount: Cardinal;
  SmoothIndices: PSmoothIndexArray;

  //--------------- local functions -------------------------------------------

  function IsVertexMarked(P: Pointer; Index: Integer): Boolean; assembler;

  // tests the Index-th bit, returns True if set else False

  asm
                     BT [EAX], EDX
                     SETC AL
  end;

  //---------------------------------------------------------------------------

  function MarkVertex(P: Pointer; Index: Integer): Boolean; assembler;

  // sets the Index-th bit and return True if it was already set else False

  asm
                     BTS [EAX], EDX
                     SETC AL
  end;

  //---------------------------------------------------------------------------

  procedure StoreSmoothIndex(ThisIndex, SmoothingGroup, NewIndex: Cardinal; P: Pointer);

  // Stores new vertex index (NewIndex) into the smooth index array of vertex ThisIndex
  // using field SmoothingGroup, which must not be 0.
  // For each vertex in the vertex array (also for duplicated vertices) an array of 32 cardinals
  // is maintained (each for one possible smoothing group. If a vertex must be duplicated because
  // it has no smoothing group or a different one then the index of the newly created vertex is
  // stored in the SmoothIndices to avoid loosing the conjunction between not yet processed vertices
  // and duplicated vertices.
  // Note: Only one smoothing must be assigned per vertex. Some available models break this rule and
  //       have more than one group assigned to a face. To make the code fail safe the group ID
  //       is scanned for the lowest bit set.

  asm
                   PUSH EBX
                   BSF EBX, EDX                  // determine smoothing group index (convert flag into an index)
                   MOV EDX, [P]                  // get address of index array
                   SHL EAX, 7                    // ThisIndex * SizeOf(TSmoothIndexEntry)
                   ADD EAX, EDX
                   LEA EDX, [4 * EBX + EAX]      // Address of array + vertex index + smoothing group index
                   MOV [EDX], ECX
                   POP EBX
  end;

  //---------------------------------------------------------------------------

  function GetSmoothIndex(ThisIndex, SmoothingGroup: Cardinal; P: Pointer): Cardinal;

  // Retrieves the vertex index for the given index and smoothing group.
  // This redirection is necessary because a vertex might have been duplicated.

  asm
                   PUSH EBX
                   BSF EBX, EDX                  // determine smoothing group index
                   SHL EAX, 7                    // ThisIndex * SizeOf(TSmoothIndexEntry)
                   ADD EAX, ECX
                   LEA ECX, [4 * EBX + EAX]      // Address of array + vertex index + smoothing group index
                   MOV EAX, [ECX]
                   POP EBX
  end;
  
  //---------------------------------------------------------------------------

  procedure DuplicateVertex(Index: Integer);

  // extends the vector and normal array by one entry and duplicates the vertex data given by Index
  // the marker and texture arrays will be extended too, if necessary

  begin
    // enhance vertex array
    ReallocMem(Result.Vertices, (CurrentVertexCount + 1) * SizeOf(TVector3f));
    Result.Vertices[CurrentVertexCount] := Result.Vertices[Index];

    // enhance normal array
    ReallocMem(Result.Normals, (CurrentVertexCount + 1) * SizeOf(TVector3f));
    Result.Normals[CurrentVertexCount] := NullVector;

    // enhance smooth index array
    ReallocMem(SmoothIndices, (CurrentVertexCount + 1) * SizeOf(TSmoothIndexEntry));
    FillChar(SmoothIndices[CurrentVertexCount], SizeOf(TSmoothIndexEntry), $FF);

    // enhance marker array
    if (CurrentVertexCount div 8) <> ((CurrentVertexCount + 1) div 8) then
    begin
      ReallocMem(Marker, ((CurrentVertexCount + 1) div 8) + 1);
      Marker[(CurrentVertexCount div 8) + 1] := 0;
    end;

    // enhance texture coordinates array
    if assigned(Result.TexCoords) then
    begin
      ReallocMem(Result.TexCoords, (CurrentVertexCount + 1) * SizeOf(TTexVert3DS));
      Result.TexCoords[CurrentVertexCount] := Result.TexCoords[Index];
    end;

    Inc(CurrentVertexCount);
  end;

  //---------------------------------------------------------------------------

  function CountBits(const Value: Cardinal): Cardinal;

  // determines number of set bits in Value
  
  asm
                   XOR EDX, EDX
                   MOV ECX, 32
  @@1:             SHR EAX, 1
                   ADC EDX, 0
                   DEC ECX
                   JNZ @@1
                   MOV EAX, EDX
  end;

  //--------------- end local functions ---------------------------------------

var
  Size: Cardinal;
  Material,
  I: Integer;
  FaceGroup: PFaceGroup;
  Vector1,
  Vector2,
  Normal: TAffineVector;
  Face,
  Vertex,
  TargetVertex: Integer;
  SmoothingGroup: Cardinal;
  CurrentIndex: Word;

  WSG: Boolean;

begin
  with Mesh3DS^ do
  begin
    if NVertices < 3 then
    begin
      // unbelievable, what terrible objects sometimes can be found
      Result := nil;
      Exit;
    end;
    // make a copy of the vertex data, this must always be available
    CurrentVertexCount := NVertices;
    // New() just calls GetMem, but I want the memory cleared
    Result := AllocMem(SizeOf(TMeshObject));
    Result.MeshName := Name;
    Result.Visible := not IsHidden;
    Size := NVertices * SizeOf(TPoint3DS);
    // allocate memory, we do not need to clear it
    GetMem(Result.Vertices, Size);
    // we don't need to consider the local mesh matrix here, since all vertices are already
    // transformed into their final positions
    Move(VertexArray^, Result.Vertices^, Size);
    {for I := 0 to NVertices - 1 do
    begin
      Result.Vertices[I, 0] := VertexArray[I].X;
      Result.Vertices[I, 1] := -VertexArray[I].Z;
      Result.Vertices[I, 2] := VertexArray[I].Y;
    end;}

    // texturing data available (optional)?
    if NTextVerts > 0 then
    begin
      Size := NTextVerts * SizeOf(TTexVert3DS);
      GetMem(Result.TexCoords, Size);
      Move(TextArray^, Result.TexCoords^, Size);
    end;

    // allocate memory for the face normal array, the final normal array and the marker array
    Result.Normals := AllocMem(NVertices * SizeOf(TVector3f));
    Marker := AllocMem((NVertices div 8) + 1); // one bit for each vertex
    GetMem(SmoothIndices, NVertices * SizeOf(TSmoothIndexEntry));

    WSG := False;

    if SmoothArray = nil then
    begin
      // no smoothing groups to consider
      for Face := 0 to NFaces - 1 do
      with FaceArray[Face], Result^ do
      begin
        // normal vector for the face
        Vector1 := VectorAffineSubtract(Vertices[V1], Vertices[V2]);
        Vector2 := VectorAffineSubtract(Vertices[V3], Vertices[V2]);
        if Version < rlRelease3 then Normal := VectorCrossProduct(Vector1, Vector2)
                                 else Normal := VectorCrossProduct(Vector2, Vector1);
        // go for each vertex in the current face
        for Vertex := 0 to 2 do
        begin
          // copy current index for faster access
          CurrentIndex := FaceRec[Vertex];
          // already been touched?
          if IsVertexMarked(Marker, CurrentIndex) then
          begin
            // already touched vertex must be duplicated
            DuplicateVertex(CurrentIndex);
            FaceRec[Vertex] := CurrentVertexCount - 1;
            Normals[CurrentVertexCount - 1] := Normal;
          end
          else
          begin
            // not yet touched, so just store the normal
            Normals[CurrentIndex] := Normal;
            MarkVertex(Marker, CurrentIndex);
          end;
        end;
      end;
    end
    else
    begin
      // smoothing groups are to be considered
      for Face := 0 to NFaces - 1 do
      with FaceArray[Face], Result^ do
      begin
        // normal vector for the face
        Vector1 := VectorAffineSubtract(Vertices[V1], Vertices[V2]);
        Vector2 := VectorAffineSubtract(Vertices[V3], Vertices[V2]);
        if Version < rlRelease3 then Normal := VectorCrossProduct(Vector1, Vector2)
                                else Normal := VectorCrossProduct(Vector2, Vector1);
        SmoothingGroup := SmoothArray[Face];
        if ShowWarnings then
        begin
          if CountBits(SmoothingGroup) > 1 then WSG := True;
        end;
        
        // go for each vertex in the current face
        for Vertex := 0 to 2 do
        begin
          // copy current index for faster access
          CurrentIndex := FaceRec[Vertex];
          // Has vertex already been touched?
          if IsVertexMarked(Marker, CurrentIndex) then
          begin
            // check smoothing group
            if SmoothingGroup = 0 then
            begin
              // no smoothing then just duplicate vertex
              DuplicateVertex(CurrentIndex);
              FaceRec[Vertex] := CurrentVertexCount - 1;
              Normals[CurrentVertexCount - 1] := Normal;
              // mark new vertex also as touched
              MarkVertex(Marker, CurrentVertexCount - 1);
            end
            else
            begin
              // this vertex must be smoothed, check if there's already a (duplicated) vertex
              // for this smoothing group
              TargetVertex := GetSmoothIndex(CurrentIndex, SmoothingGroup, SmoothIndices);
              if TargetVertex < 0 then
              begin
                // vertex has not yet been duplicated for this smoothing group, so do it now
                DuplicateVertex(CurrentIndex);
                FaceRec[Vertex] := CurrentVertexCount - 1;
                Normals[CurrentVertexCount - 1] := Normal;
                StoreSmoothIndex(CurrentIndex, SmoothingGroup, CurrentVertexCount - 1, SmoothIndices);
                StoreSmoothIndex(CurrentVertexCount - 1, SmoothingGroup, CurrentVertexCount - 1, SmoothIndices);
                // mark new vertex also as touched
                MarkVertex(Marker, CurrentVertexCount - 1);
              end
              else
              begin
                // vertex has already been duplicated, so just add normal vector to other vertex...
                Normals[TargetVertex] := VectorAffineAdd(Normals[TargetVertex], Normal);
                // ...and tell which new vertex has to be used from now on
                FaceRec[Vertex] := TargetVertex;
              end;
            end;
          end
          else
          begin
            // vertex not yet touched, so just store the normal
            Normals[CurrentIndex] := Normal;
            // initialize smooth indices for this vertex
            FillChar(SmoothIndices[CurrentIndex], SizeOf(TSmoothIndexEntry), $FF);
            if SmoothingGroup <> 0 then StoreSmoothIndex(CurrentIndex, SmoothingGroup, CurrentIndex, SmoothIndices);
            MarkVertex(Marker, CurrentIndex);
          end;
        end;
      end;
    end;
    FreeMem(Marker);
    FreeMem(SmoothIndices);

    if ShowWarnings then
    begin
      if WSG then MessageDlg('Wrong smoothing group assignments!'#13#10 +
                              'Mesh "' + Name + '" contains vertices assigned to more than one smoothing group.',
                              mtWarning, [mbOK], 0);
    end;

    // sum up all vertices to do auto centering
    Result.Count := CurrentVertexCount;
    with Result^ do
    begin
      for I := 0 to CurrentVertexCount - 1 do
        FCenter := VectorAffineAdd(FCenter, Vertices[I]);
    end;

    // finally normalize the Normals array
    with Result^ do
      for I := 0 to CurrentVertexCount - 1 do VectorNormalize(Normals[I]);

    // now go for each material group
    Result.FaceGroups := TList.Create;
    // if there's no face to material assignment then just copy the
    // face definitions and use the preset default material
    if NMats = 0 then
    begin
      FaceGroup := AllocMem(SizeOf(TFaceGroup));
      with FaceGroup^ do
      begin
        TextureIndex := 0;
        IndexCount := 3 * NFaces;
        GetMem(Indices, IndexCount * SizeOf(Integer));
        // copy the face list
        for I := 0 to NFaces - 1 do
        begin
          Indices[3 * I + 0] := FaceArray[I].V1;
          Indices[3 * I + 1] := FaceArray[I].V2;
          Indices[3 * I + 2] := FaceArray[I].V3;
        end;
      end;
      Result.FaceGroups.Add(FaceGroup);
    end
    else
    begin
      for Material := 0 to NMats - 1 do
      begin
        FaceGroup := AllocMem(SizeOf(TFaceGroup));
        with FaceGroup^ do
        begin
          if not FMaterials.Find(MatArray[Material].Name, TextureIndex) then
          begin
            ShowMessage('Material ''' + MatArray[Material].Name + ''' not found. Using default material.');
            TextureIndex := 0;
          end;
          IndexCount := 3 * MatArray[Material].NFaces;
          GetMem(Indices, IndexCount * SizeOf(Integer));
          // copy all vertices belonging to the current face into our index array,
          // there won't be redundant vertices since this would mean a face has more than one
          // material
          with MatArray[Material] do
            for I := 0 to NFaces - 1 do // NFaces is the one from FaceGroup
            begin
              Indices[3 * I + 0] := FaceArray[FaceIndex[I]].V1;
              Indices[3 * I + 1] := FaceArray[FaceIndex[I]].V2;
              Indices[3 * I + 2] := FaceArray[FaceIndex[I]].V3;
            end;
        end;
        Result.FaceGroups.Add(FaceGroup);
      end;
    end;
  end;
end;
{$ifdef RangeCheck} {$undef RangeCheck} {$R+} {$endif}

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ClearLists;

// free locally allocated data and reset structures, variables etc.

var I, J: Integer;

begin
  if FMainList > 0 then
  begin
    glDeleteLists(FMainList, 1);
    FMainList := 0;
  end;

  for I := 0 to FLists.Count - 1 do
    with PMeshObject(FLists[I])^ do
    begin
      glDeleteLists(ListName, 1);
      if Assigned(Vertices) then  FreeMem(Vertices);
      if Assigned(Normals) then FreeMem(Normals);
      if Assigned(TexCoords) then FreeMem(TexCoords);
      // free material assignment data
      for J := 0 to FaceGroups.Count - 1 do
      begin
        FreeMem(PFaceGroup(FaceGroups[J]).Indices);
        FreeMem(FaceGroups[J]);
      end;
      FaceGroups.Free;
      FreeMem(FLists[I]);
    end;
  FLists.Clear;

  for I := 0 to FMaterials.Count - 1 do
  begin
    if PMaterial(FMaterials.Objects[I]).TextureObject <> 0 then
      glDeleteTextures(1, @PMaterial(FMaterials.Objects[I]).TextureObject);
    FreeMem(Pointer(FMaterials.Objects[I]));
  end;
  FMaterials.Clear;

  for I := 0 to FLights.Count - 1 do
  begin
    glDisable(GL_LIGHT0 + I);
    FreeMem(Pointer(FLights.Objects[I]));
  end;
  FLights.Clear;

  if FBackgroundTexture <> 0 then glDeleteTextures(1, @FBackgroundTexture);
  FBackgroundTexture := 0;
  FFocalLength := 70;
  FLLabel.Caption := Format('focal length: %d mm', [FFocalLength]);
  FDepthOfView := 10000;
  FCameraChanged := True;
  FSceneMatrix := IdentityMatrix;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.Use3DSData(const Reader: TFile3DS);

// Notification method used by the main form to tell the viewer to upate its display.
// Previously loaded data will be freed and, if Reader is <> nil, new data will be loaded.
// Note: The rendering context must be active because display lists will be created herein.
//       All converted data (vertices, normals etc.) is kept until a new file is to be loaded
//       or the viewer is freed. 

  //--------------- local functions -------------------------------------------

  function FindMeshIndex(const ObjectName: String): Integer;

  // Searchs through the list of defined objects and returns the index
  // corresponding to the given name or -1 if the object is not defined. Due to the
  // way instances of meshs are stored the name might be of the form "Mesh.Instance".
  // This case is handled as would there only be a mesh name.
  // NOTE: name matching is not case sensitive

  var I: Integer;
      Name: String;

  begin
    Result := -1;
    if Pos('.', ObjectName) > 0 then Name := Copy(ObjectName, 1, Pos('.', ObjectName) - 1)
                                else Name := ObjectName;
    for I := 0 to FLists.Count - 1 do
      with PMeshObject(FLists[I])^ do
        if CompareText(MeshName, Name) = 0 then
        begin
          Result := I;
          Break;
        end;
  end;

  //---------------------------------------------------------------------------

  function FindMotionIndex(const ObjectName: String): Integer;

  // Looks through the motion list for the object "ObjectName" and returns its index
  // or -1 if the name is not it the list

  var I: Integer;

  begin
    Result := -1;
    with Reader.KeyFramer do
      for I := 0 to MeshMotionCount - 1 do
        if CompareText(MeshMotion[I].Name, ObjectName) = 0 then
        begin
          Result := I;
          Break;
        end;
  end;

  //---------------------------------------------------------------------------

  function InvertMeshMatrix(const Name: String): TMatrix;

  // constructs a 4x4 matrix from 3x4 local mesh matrix given by Name and
  // inverts it so it can be used for the keyframer stuff

  var I, Index: Integer;
  
  begin
    with Reader.Objects do
    begin
      Index := -1;
      for I := 0 to MeshCount - 1 do
        if CompareText(Mesh[I].Name, Name) = 0 then
        begin
          Index := I;
          Break;
        end;

      if Index > -1 then
      begin
        with Mesh[Index]^ do
        begin
          Result[0, 0] := LocMatrix[0];
          Result[0, 1] := LocMatrix[1];
          Result[0, 2] := LocMatrix[2];
          Result[0, 3] := 0;
          Result[1, 0] := LocMatrix[3];
          Result[1, 1] := LocMatrix[4];
          Result[1, 2] := LocMatrix[5];
          Result[1, 3] := 0;
          Result[2, 0] := LocMatrix[6];
          Result[2, 1] := LocMatrix[7];
          Result[2, 2] := LocMatrix[8];
          Result[2, 3] := 0;
          Result[3, 0] := LocMatrix[9];
          Result[3, 1] := LocMatrix[10];
          Result[3, 2] := LocMatrix[11];
          Result[3, 3] := 1;
        end;
        MatrixInvert(Result);
      end
      else Result := IdentityMatrix;
    end;
  end;

  //--------------- end local functions ---------------------------------------

var I, J: Integer;
    Mesh: PMeshObject;
    Release: TReleaseLevel;
    TotalCount: Cardinal;
//    CurrentFrameFlags: Word;
    ParentMotionIndex,
    CurrentMeshIndex: Integer;
    Matrix,
    Matrix2: TMatrix;
    Picture: TPicture;
    TargetWidth,
    TargetHeight: Integer;
    MemDC: HDC; 

begin
  // clear previously allocated data
  ProgressMessage('uploading to OpenGL: clearing old data...');
  ClearLists;
  if Assigned(Reader) and (Reader.Objects.MeshCount > 0) then
  begin
    // make sure we have a valid release level
    Release := Reader.MeshRelease;
    // database release overrides mesh release
    if (Release = rlReleaseNotKnown) or (Release < Reader.DatabaseRelease)  then Release := Reader.DatabaseRelease;
    if Release = rlReleaseNotKnown then Release := rlRelease3;

    // copy background settings
    ProgressMessage('uploading to OpenGL: loading background...');
    FBackground := Reader.Background;
    with FBackground do
    begin
      if BgndUsed = btNoBgnd then
      begin
        // make default background if no other is given
        BgndUsed := btUseSolidBgnd;
        Solid.R := 0;
        Solid.G := 0;
        Solid.B := 0;
      end
      else
        if BgndUsed = btUseBitmapBgnd then
        begin
          MemDC := CreateCompatibleDC(0);
          Picture := TPicture.Create;
          if LoadPictureFromFile(Picture, ExtractFilePath(Reader.FileName), Bitmap) then
          begin
            // check width and height being power of 2
            with Picture do
            begin
              TargetWidth := RoundDownToPowerOf2(Width);
              TargetHeight := RoundDownToPowerOf2(Height);
              // stretch image if necessary
              if (TargetWidth <> Width) or (TargetHeight <> Height) then
                Stretch(TargetWidth, TargetHeight, sfLanczos3, 0, Bitmap);
              // create texture object
              FBackgroundTexture := CreateTextureObject(MemDC, Bitmap, GL_NEAREST, GL_NEAREST);
              if FBackgroundTexture = 0 then
              begin
                MessageDlg('Texture object could not be created (' + FBackground.Bitmap + ')', mtInformation, [mbOK], 0);
              end
              else
              begin
                glBindTexture(GL_TEXTURE_2D, FBackgroundTexture);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
              end;
            end;
          end;
          Picture.Free;
          DeleteDC(MemDC);
          if FBackgroundTexture = 0 then
          begin
            // make default background if bitmap background could not be loaded
            BgndUsed := btUseSolidBgnd;
            Solid.R := 0;
            Solid.G := 0;
            Solid.B := 0;
          end
        end;
      if BgndUsed = btUseSolidBgnd then glClearColor(Solid.R, Solid.G, Solid.B, 1);
    end;

    // fog settings
    if Reader.Atmosphere.ActiveAtmo in [atUseFog, atUseLayerFog] then
      with Reader.Atmosphere do
      begin
        glEnable(GL_FOG);
        glFogi(GL_FOG_MODE, GL_LINEAR);
        if ActiveAtmo = atUseFog then
        begin
          glFogfv(GL_FOG_COLOR, @Fog.FogColor);
          glFogf(GL_FOG_DENSITY, (Fog.NearDensity + Fog.FarDensity) / 2);
          glFogf(GL_FOG_START, Fog.NearPlane);
          glFogf(GL_FOG_END, Fog.FarPlane);
        end
        else
        begin
          glFogfv(GL_FOG_COLOR, @Fog.FogColor);
          glFogf(GL_FOG_DENSITY, LayerFog.Density);
          glFogf(GL_FOG_START, LayerFog.ZMin);
          glFogf(GL_FOG_END, LayerFog.ZMax);
        end
      end
      else glDisable(GL_FOG);

    ProgressMessage('uploading to OpenGL: converting materials...');
    ConvertMaterials(Reader.Materials);
    ProgressMessage('uploading to OpenGL: loading textures...');
    SetupTextures(ExtractFilePath(Reader.FileName));
    // prepare mesh structure
    FLists.Capacity := Reader.Objects.MeshCount;
    // create display lists
    FCenter := NullVector;
    TotalCount := 0;
    ProgressMessage('uploading to OpenGL: converting meshs...');
    for I := 0 to Reader.Objects.MeshCount - 1 do
    begin
      // allocate new mesh structure and fill it with mesh data
      Mesh := ConvertMeshData(Release, Reader.Objects.Mesh[I], False);
      // meshs without vertices or containing errors don't get into our list
      if Assigned(Mesh) then
      begin
        Inc(TotalCount, Mesh.Count);
        FLists.Add(Mesh);
      end;
    end;
    // FCenter is updated during conversion of the meshs
    VectorScale(FCenter, 1 / TotalCount);

    ProgressMessage('uploading to OpenGL: sending vertex data...');
    for I := 0 to FLists.Count - 1 do
    begin
      Mesh := FLists[I];
      with Mesh^ do
      begin
        // allocate a new display list
        ListName := glGenLists(1);
        // now compile mesh data into this list
        glNewList(ListName, GL_COMPILE);
        if Release < rlRelease3 then glFrontFace(GL_CW)
                                else glFrontFace(GL_CCW);

        glVertexPointer(3, GL_FLOAT, 0, Vertices);
        glEnableClientState(GL_VERTEX_ARRAY);

        // enable the normal array if available
        if assigned(Normals) then
        begin
          glNormalPointer(GL_FLOAT, 0, Normals);
          glEnableClientState(GL_NORMAL_ARRAY);
        end
        else glDisableCLientState(GL_NORMAL_ARRAY);

        // enable the texture coordinates array if available
        if Assigned(TexCoords) then
        begin
          glTexCoordPointer(2, GL_FLOAT, 0, TexCoords);
          glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        end
        else glDisableCLientState(GL_TEXTURE_COORD_ARRAY);

        for J := 0 to FaceGroups.Count - 1 do
          with PFaceGroup(FaceGroups[J])^ do
          begin
            // apply materials
            with PMaterial(FMaterials.Objects[TextureIndex])^ do
            begin
              glMateriali(GL_FRONT_AND_BACK, GL_SHININESS, Shininess);
              glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, @Emission);
              glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, @Ambient);
              glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, @Diffuse);
              glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, @Specular);

              glPolygonMode(GL_FRONT_AND_BACK, PolygonMode);
              glShadeModel(ShadeModel);
              if TwoSided then glDisable(GL_CULL_FACE)
                          else glEnable(GL_CULL_FACE);

              // texturing
              if TextureObject <> 0 then
              begin
                glBindTexture(GL_TEXTURE_2D, TextureObject);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, WrapS);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, WrapT);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, MinFilter);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, MagFilter);
                glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, Mode);

                glEnable(GL_TEXTURE_2D);
              end
              else glDisable(GL_TEXTURE_2D);

              // ambient, diffuse and specular colors values contain the same blend (alpha) factor
              if Ambient[3] > 0 then
              begin
                glEnable(GL_BLEND);
                // depth buffer should be disabled while rendering semi transparent surfaces but
                // leads often to somewhat strange results as the mesh aren't sorted properly
                //glDepthMask(GL_FALSE);
                glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);
              end
              else
              begin
                glDisable(GL_BLEND);
                //glDepthMask(GL_TRUE);
              end;
            end;
            // now draw the mesh data
            glDrawElements(GL_TRIANGLES, IndexCount, GL_UNSIGNED_INT, Indices);
          end;
      
        // finished compiling this display list
        glEndList;
        CheckOpenGLError;
      end;
    end;

    ProgressMessage('uploading to OpenGL: converting light sources...');
    ConvertLightSources(Reader.Objects);
    
    // build main display list
    FMainList := glGenLists(1);
    glNewList(FMainList, GL_COMPILE);
    // move scene so that its geometric center is at origin
    glTranslatef(-FCenter[0], -FCenter[1], -FCenter[2]);

    // if no keyframer data is available then just display all defined objects
    if Reader.KeyFramer.MeshMotionCount = 0 then
    begin
      for I := 0 to FLists.Count - 1 do
        with PMeshObject(FLists[I])^ do
          if Visible then glCallList(ListName);
    end
    else
      ProgressMessage('uploading to OpenGL: converting keyframer data...');
      with Reader.Keyframer do
      begin
        // use keyframer data (frame 0 only)
        for I := 0 to MeshMotionCount - 1 do
        begin
          //CurrentFrameFlags := MeshMotion[I].Flags1;
          // get index of current object in the mesh list...
          CurrentMeshIndex := FindMeshIndex(MeshMotion[I].Name);
          // ...as well as its parent object if there's one
          ParentMotionIndex := FindMotionIndex(MeshMotion[I].Parent);
          // mesh might not be defined for dummy objects used in the keyframer
          if CurrentMeshIndex > -1 then
          begin
            // if there's no parent then we don't need to reapply all transformations
            // as all vertices of a mesh are already at the correct place
            if ParentMotionIndex > -1 then
            begin
              // convert keyframer data into a transformation matrix (frame 0 only);
              // mesh vertices are already at their final places, we need to revert them to their
              // original positions to apply further transformations
              Matrix := InvertMeshMatrix(MeshMotion[I].Name);
              // consider pivot point (translation simply made...)
              Matrix[3, 0] := Matrix[3, 0] - MeshMotion[I].Pivot.X;
              Matrix[3, 1] := Matrix[3, 1] - MeshMotion[I].Pivot.Y;
              Matrix[3, 2] := Matrix[3, 2] - MeshMotion[I].Pivot.Z;

              // scaling
              if (MeshMotion[I].NSKeys > 0) and (MeshMotion[I].SKeys[0].Time = 0) then
              begin
                Matrix2 := CreateScaleMatrix(TAffineVector(MeshMotion[I].Scale[0]));
                Matrix := MatrixMultiply(Matrix, Matrix2);
              end;

              // rotation
              if (MeshMotion[I].NRKeys > 0) and (MeshMotion[I].RKeys[0].Time = 0) then
              begin
                with MeshMotion[I].Rot[0] do
                  Matrix2 := CreateRotationMatrix(MakeAffineVector([X, Y, Z]), Angle);
                Matrix := MatrixMultiply(Matrix, Matrix2);
              end;

              // translation
              if (MeshMotion[I].NPKeys > 0) and (MeshMotion[I].PKeys[0].Time = 0) then
              begin
                Matrix[3, 0] := Matrix[3, 0] + MeshMotion[I].Pos[0].X;
                Matrix[3, 1] := Matrix[3, 1] + MeshMotion[I].Pos[0].Y;
                Matrix[3, 2] := Matrix[3, 2] + MeshMotion[I].Pos[0].Z;
              end;

              // now apply transformations of all of the mesh's parents
              while ParentMotionIndex > -1 do
              begin
                // scaling
                if (MeshMotion[ParentMotionIndex].NSKeys > 0) and (MeshMotion[ParentMotionIndex].SKeys[0].Time = 0) then
                begin
                  Matrix2 := CreateScaleMatrix(TAffineVector(MeshMotion[ParentMotionIndex].Scale[0]));
                  Matrix := MatrixMultiply(Matrix, Matrix2);
                end;

                // rotation
                if (MeshMotion[ParentMotionIndex].NRKeys > 0) and (MeshMotion[ParentMotionIndex].RKeys[0].Time = 0) then
                begin
                  with MeshMotion[ParentMotionIndex].Rot[0] do
                    Matrix2 := CreateRotationMatrix(MakeAffineVector([X, Y, Z]), Angle);
                  Matrix := MatrixMultiply(Matrix, Matrix2);
                end;

                // translation
                if (MeshMotion[ParentMotionIndex].NPKeys > 0) and (MeshMotion[ParentMotionIndex].PKeys[0].Time = 0) then
                begin
                  Matrix[3, 0] := Matrix[3, 0] + MeshMotion[ParentMotionIndex].Pos[0].X;
                  Matrix[3, 1] := Matrix[3, 1] + MeshMotion[ParentMotionIndex].Pos[0].Y;
                  Matrix[3, 2] := Matrix[3, 2] + MeshMotion[ParentMotionIndex].Pos[0].Z;
                end;

                ParentMotionIndex := FindMotionIndex(MeshMotion[ParentMotionIndex].Parent);
              end;

              glPushMatrix;
              glMultMatrixf(@Matrix);
              with PMeshObject(FLists[CurrentMeshIndex])^ do
                if Visible then glCallList(ListName);
              glPopMatrix;
            end
            else
              with PMeshObject(FLists[CurrentMeshIndex])^ do
                if Visible then glCallList(ListName);
          end;
        end;
      end;
    glEndList;
    CheckOpenGLError;
  end;
  Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.CheckOpenGLError;

// Gets the oldest error from OpenGL engine and tries to clear the error queue.
// Because under some circumstances reading the error code creates a new error
// and thus hanging up the thread, we limit the loop to 6 reads.

var GLError: TGLEnum;
    Count: Word;

begin
  GLError := glGetError;
  if GLError <> GL_NO_ERROR then
  begin
    Count := 0;
    while (glGetError <> GL_NO_ERROR) and (Count < 6) do Inc(Count);
    raise EOpenGLError.Create(gluErrorString(GLError));
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.CreateParams(var Params: TCreateParams);

begin
  inherited;
  with Params do
    // By using the CS_OWNDC flag we tell Windows not to reset all our settings every time
    // we access the window's DC. This enables to activate our rendering context just once (when the
    // window is created) and not every time a paint cycle occurs.
    // The redraw flags are necessary because the window needs a repaint on size change (viewport
    // depends on window size).
    WindowClass.Style := WindowClass.Style or CS_OWNDC or CS_VREDRAW or CS_HREDRAW;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ColorClick(Sender: TObject);

var Color: Cardinal;

begin
  Color := ColorToRGB(StringToColor(TMenuItem(Sender).Caption));
  with FBackground do
  begin
    BgndUsed := btUseSolidBgnd;
    Solid.R := GetRValue(Color) / 255;
    Solid.G := GetGValue(Color) / 255;
    Solid.B := GetBValue(Color) / 255;
    glClearColor(Solid.R, Solid.G, Solid.B, 1);
  end;
  Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.EnumColor(const ColorName: String);

begin
  BkgndItem.Add(NewItem(ColorName, 0, False, True, ColorClick, 0, ColorName + 'Item'));
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);

// store mouse down position for navigation

begin
  SetFocus;
  FLastX := X;
  FLastY := Y;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);

// calculate scene rotation values

begin
  if ssLeft in Shift then
  begin
    // This seems a bit more complex than it needs to be but I want to rotate the scene around the
    // the gobal X and Y axes (world axes) not the local ones of the scene.
    FSceneMatrix := Pitch(FSceneMatrix, XVector, DegToRad(FLastY - Y));
    if ssCtrl in Shift then FSceneMatrix := Roll(FSceneMatrix, ZVector, DegToRad(X - FLastX))
                       else FSceneMatrix := Turn(FSceneMatrix, YVector, DegToRad(FLastX - X));

    FLastX := X;
    FLastY := Y;
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ScaleUpDownChanging(Sender: TObject; var AllowChange: Boolean);

begin
  Invalidate;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ScaleEditKeyPress(Sender: TObject; var Key: Char);

begin
  if (Key = #13) then
  begin
    StrToInt(ScaleEdit.Text);
    FLightSourcesChanged := True;
    Invalidate;
    Key := #0;
  end;
  if Key in ['A'..'Z', 'a'..'z', '+', '-'] then Key := #0;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.NavCenterItemClick(Sender: TObject);

begin
  with NavCenterItem do
  begin
    Checked := not Checked;
    NavPanel.Visible := Checked;
    FormResize(nil);
    Invalidate;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);

begin
  case Key of
    VK_ADD:
      begin
        Inc(FFocalLength);
        FLLabel.Caption := Format('focal length: %d mm', [FFocalLength]);
        FLLabel.Update;
        FCameraChanged := True;
        Key := 0;
        Invalidate;
      end;
    VK_SUBTRACT:
      begin
        Dec(FFocalLength);
        FLLabel.Caption := Format('focal length: %d mm', [FFocalLength]);
        FLLabel.Update;
        FCameraChanged := True;
        Key := 0;
        Invalidate;
      end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

procedure TViewerForm.ConvertLightSources(const Objects: TObjectList);

// Builds an internal list with values which can directly sent to OpenGL.

var Light: PLightSource;
    I: Integer;
    
begin
  FLightSourcesChanged := True;
  with Objects do
  begin
    for I := 0 to OmniLightCount - 1 do
    begin
      Light := AllocMem(SizeOf(TLightSource));
      FLights.AddObject(OmniLight[I].Name, Pointer(Light));
      with Light^, OmniLight[I]^ do
      begin
        Enabled := not DLOff;
        IsSpot := False;
        Position := MakeVector([Pos.X, Pos.Y, Pos.Z, 1]);
        // we have only one color (don't know exactly how to distribute)
        Ambient := MakeVector([Color.R / 5, Color.G / 5, Color.B / 5, 1]);
        Diffuse := MakeVector([Color.R, Color.G, Color.B, 1]);
        Specular := MakeVector([Color.R, Color.G, Color.B, 1]);
        // needed to reset spot characteristic if it is was enabled before
        SpotCutOff := 180;
      end;
    end;

    for I := 0 to SpotLightCount - 1 do
    begin
      Light := AllocMem(SizeOf(TLightSource));
      FLights.AddObject(SpotLight[I].Name, Pointer(Light));
      with Light^, SpotLight[I]^ do
      begin
        Enabled := not DLOff;
        IsSpot := True;
        Position := MakeVector([Pos.X, Pos.Y, Pos.Z, 1]);
        // we have only one color (don't know exactly how to distribute)
        Ambient := MakeVector([Color.R / 5, Color.G / 5, Color.B / 5, 1]);
        Diffuse := MakeVector([Color.R, Color.G, Color.B, 1]);
        Specular := MakeVector([Color.R, Color.G, Color.B, 1]);

        with Spot^ do
        begin
          SpotDirection := MakeAffineVector([Target.X - Pos.X, Target.Y - Pos.Y, Target.Z - Pos.Z]);
          SpotExponent := 0; // can the HotSpot angle be used here? 
          SpotCutOff := FallOff;
        end;
      end;
    end;
  end;
end;

//----------------------------------------------------------------------------------------------------------------------

end.
